(*<*)
theory ISM_package
imports dAmbISM

begin
ML {*
(*  Title:      ISM_package.ML
    ID:         $Id: ISM_package.ML,v 1.13 2003/09/12 16:13:02 dvo Exp $
    Author:     Sebastian Nanz, TU Muenchen and David von Oheimb, Siemens AG

Definition of Interacting State Machines (ISMs).
*)

datatype ('a,'b) sum = Inl of 'a | Inr of 'b;

fun option_map f  NONE    = NONE
|   option_map f (SOME x) = SOME (f x)

type ind_def = {defs: thm list, mono: thm, elims: thm list, intrs: thm list,
            induct: thm, unfold: thm, mk_cases: string -> thm, raw_induct: thm};


signature ISM_PACKAGE =
sig
  type 'a rule
  val add_ism: string -> (string * string) list -> 
               string -> string -> string -> string -> 
               (string * string option) option -> 
               string option -> 
               (string * (string * string)) option -> 
               (string * (string * string)) option ->
               string rule list -> theory -> 
               theory * ind_def list * thm list
  val add_ism_i: string -> (string * typ) list -> 
               typ -> term -> term -> typ -> 
               (typ * term option) option ->
               typ option -> 
               (typ * (term * string)) option -> 
               (typ * (term * term  )) option ->
               term rule list -> theory -> 
               theory * ind_def list * thm list
end;


structure ISM_package (*: ISM_PACKAGE*) =
struct

(*** utilities ***)

(** messages and errors **)

val quiet_mode = ref false;
fun message s = if !quiet_mode then () else writeln s;

exception ism_exception;
fun ism_error msg = (raise ism_exception) handle ism_exception => error msg;

fun no_state_error name = "ISM " ^ quote name ^ 
                         " has neither a control nor a data state declaration.";
fun no_ctrl_st_error name = "ISM has a control state transition in rule " ^ 
                            quote name ^ " but no control state declaration.";
fun no_ctrl_trans_error name = 
        "ISM has control state but not control state transitions in rule " ^ 
        quote name ^ ".";
fun no_data_st_error name = "ISM has postconditions in transition " ^ 
                            quote name ^ " but no data state declaration.";
fun field_type_error field_name = "Unable to determine field type for field " ^
                                  quote field_name ^ ".";
fun fetch_constr_error l_st_name g_st_name = "Unable to determine a datatype " ^
                   quote g_st_name ^ "\n that has a single argument of type " ^
                   quote l_st_name ^ ".";

(** constructor of global type **)

fun fetch_constr'' l_st_Typ constr_opt [] = constr_opt
  | fetch_constr'' l_st_Typ constr_opt (constr::rest) =
    let
      val (constr_Term, constr_Typ) = constr;
      val l_st_DTyp = DatatypeAux.dtyp_of_typ [] l_st_Typ;
      val constr_opt' = if constr_Typ = [l_st_DTyp] then SOME constr_Term 
                                                    else constr_opt;
    in
      fetch_constr'' l_st_Typ constr_opt' rest
    end;

fun fetch_constr' l_st_Typ constr_opt [] = constr_opt
  | fetch_constr' l_st_Typ constr_opt (descr::rest) =
    let
      val (_, (_, _, constr)) = descr;
      val constr_opt' = fetch_constr'' l_st_Typ constr_opt constr;
    in
      fetch_constr' l_st_Typ constr_opt' rest
    end;

fun fetch_constr l_st_Typ g_st_Typ thy =
  let
    val sign = Theory.sign_of thy;
    val fetch_constr_err = fetch_constr_error (Sign.string_of_typ sign l_st_Typ)
      (Sign.string_of_typ sign g_st_Typ);

    (* determine type name of global state datatype *)
    val g_st_name = case g_st_Typ of Type (g_st_name, _)
                         => Sign.intern_type sign g_st_name
                     | _ => let val _ = ism_error fetch_constr_err in "" end;
    
    (* symbol table lookup *)
    val symtab_opt = Symtab.lookup (DatatypePackage.get_datatypes thy) g_st_name;
    val descr_opt = option_map (#descr) symtab_opt;
    val descr = case descr_opt of SOME descr => descr
      | NONE => let val _ = ism_error fetch_constr_err in [] end;
    val constr_opt = fetch_constr' l_st_Typ NONE descr;
    val constr = case constr_opt of NONE =>
      let val _ = ism_error fetch_constr_err in "" end
      | SOME(constr) => constr;
  in
    constr
  end;

fun mk_global_st_term l_st_Typ g_st_Typ_opt thy t =
  case g_st_Typ_opt of NONE           => t 
                     | SOME(g_st_Typ) => 
    Const (fetch_constr l_st_Typ g_st_Typ thy, l_st_Typ --> g_st_Typ) $ t

(* end; *)

(** helpers **)

fun combine [] [] = []
  | combine (x::xs) [] = []
  | combine [] (y::ys) = []
  | combine (x::xs) (y::ys) = (x, y) :: combine xs ys;

local
  fun esc_underscores [] = []
  |   esc_underscores (x::xs) = if (x = "_") then "'"::x::esc_underscores xs
                                             else      x::esc_underscores xs;
in
  fun escape_underscores name = implode (esc_underscores (explode name))
end;


(*** type handling ***)

type 'a rule = ((bstring * Args.src list) * (('a * 'a) option * 
               ('a list * ((('a * bool) * 'a) list * ((('a * bool) * 'a) list *
               ('a option * ((string * 'a) list, 'a) sum))))));

val dummyT = Type ("_dummy", []);

fun mk_Typ prep_typ sign type_constr = typ_of (prep_typ sign type_constr);
fun mk_listT T = Type ("List.list", [T]);



(*** term handling ***)

val dummy_Term = Const("_dummy", dummyT);

fun mk_Term prep_term sign t = term_of (prep_term sign t);

fun mk_k_nil_Term sign pt_Typ msg_Typ =
  let
    val msgss_Typ = mk_listT msg_Typ;
    val k_nil_Typ = pt_Typ --> msgss_Typ;
    val k_nil_Term = Const(Sign.intern_const sign "K_Nil", k_nil_Typ);
  in
    (k_nil_Term, k_nil_Typ)
  end;

fun mk_pair_Const sign a_Typ b_Typ =
  let
    val pair_Const_Typ = a_Typ --> b_Typ --> HOLogic.mk_prodT (a_Typ, b_Typ);
    val pair_Const = Const("Pair", pair_Const_Typ);
  in
    (pair_Const, pair_Const_Typ)
  end;

fun mk_pair_Term_Typ sign (a_Term, a_Typ) (b_Term, b_Typ) =
  let
    val (pair_Const, _) = mk_pair_Const sign a_Typ b_Typ;
    val pair_Term = pair_Const $ a_Term $ b_Term;
    val pair_Typ = HOLogic.mk_prodT(a_Typ, b_Typ);
  in
    (pair_Term, pair_Typ)
  end;

fun mk_state_opt prep_type prep_term sign = option_map
   (fn (s_type, (s_term, s_Var)) =>
    let
      val s_Typ  = mk_Typ  prep_type sign s_type;
      val s_Term = mk_Term prep_term sign (s_term, s_Typ);
    in
      (s_Term, s_Typ, s_Var)
    end);



(*** transitions ***)

(** prepare inputs and outputs **)

fun mk_msgss_pat sign pt_Typ msg_Typ = 
  let
    val msg_list_Typ = mk_listT msg_Typ;
    val msgss_Typ = pt_Typ --> msg_list_Typ;
    fun fun_upd_Typ T = msgss_Typ --> T --> msg_list_Typ --> msgss_Typ;
    val fun_upd_Term   = Const (Sign.intern_const sign "fun_upd" , 
                                fun_upd_Typ pt_Typ);
    val fun_upd_s_Term = Const (Sign.intern_const sign "fun_upd_s", 
                                fun_upd_Typ (HOLogic.mk_setT pt_Typ));
   in
      Library.foldl (fn (f, ((pat,multi), msgs)) => 
          (if multi then fun_upd_s_Term else fun_upd_Term) $ f $ pat $ msgs)
    end;


(** update record fields (for postconditions) **)
fun find_field_Typ sign field_name data_st_Typ =
  case data_st_Typ of (Type (field_name', [field_Typ, record_Typ])) =>
    if ((Sign.intern_const sign (field_name ^ "_ext_type")) = field_name')
       then field_Typ else find_field_Typ sign field_name record_Typ
  | _ => let val _ = ism_error (field_type_error field_name) in dummyT end;

(* Typ: Context.theory -> Term.typ -> Term.term -> (string * Term.term) list -> Term.term *)
fun mk_field_upd sign data_st_Typ r [] = r
  | mk_field_upd sign data_st_Typ r (post::rest) =
    let
      val (field_name, field_Term) = post;
      val field_Typ = find_field_Typ sign field_name data_st_Typ;
      val field = Const(Sign.intern_const sign (field_name ^ "_update"),
        field_Typ --> data_st_Typ --> data_st_Typ) $ field_Term $ r;
    in
      mk_field_upd sign data_st_Typ field rest
    end;


(** add premises terms to conclusion term **)

fun mk_prems sign concl [] = concl
  | mk_prems sign concl (pre_Term::rest) =
    let
      val impl_Term = Const("==>", propT --> propT --> propT);
      val concl' = impl_Term $ (HOLogic.mk_Trueprop pre_Term) $ concl;
    in
      mk_prems sign concl' rest
    end;

(** transform transition rule (for simultaneous type check) **)

(* Typ: (Context.theory -> (''a * Term.typ) list -> Thm.cterm list) ->
       Context.theory ->
       (Term.typ * 'b) option * Term.typ * Term.typ * 'c *
       ('d * ...) option * ... option * ... ->
       ''a ISM_package.rule -> Term.term option * Term.term ISM_package.rule
*)
fun transform_rule prep_rule sign types (rule:''a rule) =
  let
    val (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ, ctrl_st_Typ_Term_opt,
	      data_st_Typ_Term_opt, l_st_Typ, g_st_Typ_opt) = types;
    val ((name, args), (cs_exprs, (pres, (inps, (outps, (cmd_opt, post_v)))))) =
        rule;
    val msg_list_Typ = mk_listT msg_Typ;
    val ctrl_st_Typ = case ctrl_st_Typ_Term_opt of 
                        NONE => (case cs_exprs of NONE => dummyT 
                                  | SOME _ => ism_error (no_ctrl_st_error name))
                       | SOME (_, typ, _) => (case cs_exprs of NONE => 
                          ism_error (no_ctrl_trans_error name) | SOME _ => typ);

    (* transform rule into Term_Typ list (for prep_rule) *)
    val l = case data_st_Typ_Term_opt of NONE => []
      | SOME(_, data_st_Typ, data_st_var) => [(data_st_var, data_st_Typ)];
    val l = l @ (case cs_exprs of NONE => [] | SOME (cs_expr,cs_expr') =>
                 [(cs_expr, ctrl_st_Typ), (cs_expr', ctrl_st_Typ)]);
    val l = l @ (map (fn p => (p, HOLogic.boolT)) pres);
    val trans_forw_pats = Library.flat o map (fn ((pat, multi), msgs) =>
                       [(pat, if multi then HOLogic.mk_setT pt_Typ else pt_Typ),
                        (msgs, msg_list_Typ)]);
    val l = l @ trans_forw_pats inps; 
    val l = l @ trans_forw_pats outps;
    val l = l @ (case cmd_opt of NONE => []
    | SOME cmd => (case cmd_Typ_default_opt of
                     NONE => ism_error ("ISM has no commands " ^ 
                      "but subsection \"cmd\" is present in rule " ^quote name)
                   | SOME (cmd_Typ, _) => [(cmd, cmd_Typ)]));
    val l = case data_st_Typ_Term_opt of NONE => (if (post_v <> Inl []) then 
                                   ism_error (no_data_st_error name) else l)
                              | SOME (_, data_st_Typ, _) => l @ (case post_v of 
          Inl ps => (map (fn (f,v) => (v,find_field_Typ sign f data_st_Typ))) ps
        | Inr sval => [(sval, data_st_Typ)]);

    (* transformation *)
    val l = map term_of (prep_rule sign l);

    (* transform Term_Typ list back into rule *)
    val (data_st_var_opt,l) = case data_st_Typ_Term_opt of NONE => (NONE,l)
                              | _ => (SOME (hd l), Library.drop (1, l));
    val (ctrl_st_exprs,l) = (case cs_exprs of NONE => (NONE,l) 
                             | SOME _ => (SOME (hd l, hd (tl l)), Library.drop(2, l)));
    val pres' = Library.take (length pres, l);
    val l = Library.drop (length pres, l);
    fun trans_back_pats l [] = []
    |   trans_back_pats l (((pat,multi),_)::rest) = 
        ((hd l, multi), hd (tl l))::trans_back_pats (tl (tl l)) rest;
    val inps' = trans_back_pats l inps;
    val l = Library.drop (2* length inps, l);
    val outps' = trans_back_pats l outps;
    val l = Library.drop (2* length outps, l);
    val (cmd_opt',l) = if cmd_opt = NONE then (NONE,l) 
                       else (SOME (hd l), Library.drop (1, l));
    val post_v' = case post_v of 
          Inl posts => Inl (combine ((map (fn (f, v) => f)) posts) l)
        | Inr sval =>  Inr (hd l);
    val rule' = ((name, args), 
                (ctrl_st_exprs, (pres', (inps', (outps',(cmd_opt',post_v'))))));
  in
    (data_st_var_opt, (rule':term rule))
  end;


(** make introduction rules **)

fun mk_intros prep_rule thy types param_transs_Term rule =
    let
      val sign = Theory.sign_of thy;

      (* expand types *)
      val (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ, ctrl_st_Typ_Term_opt,
        data_st_Typ_Term_opt, l_st_Typ, g_st_Typ_opt) = types;

      val (data_st_var_opt, rule') = 
               transform_rule prep_rule sign types rule;

      (* expand transition *)
      val ((name, args), (ctrl_st_exprs, conds)) = rule';
      val (pres, (inps, (outps, (cmd_opt, post_v)))) = conds;

      val (k_nil_Term, _) = mk_k_nil_Term sign pt_Typ msg_Typ;
      val msgss_Typ = pt_Typ --> mk_listT msg_Typ;

      (* prepare inputs *)
      val is_Term = mk_msgss_pat sign pt_Typ msg_Typ (k_nil_Term, inps);

      (* prepare outputs *)
      val os_Term = mk_msgss_pat sign pt_Typ msg_Typ (k_nil_Term, outps);

      (* prepare states *)
      fun mk_state proj post_v = mk_global_st_term l_st_Typ g_st_Typ_opt thy 
        (case data_st_Typ_Term_opt of
          NONE => proj (the ctrl_st_exprs)
        | SOME(_, data_st_Typ, _) => 
          let
            val data_st_Term = case post_v of 
                    Inl posts => mk_field_upd sign data_st_Typ 
                                              (the data_st_var_opt) posts
                  | Inr sval => sval;
          in case ctrl_st_Typ_Term_opt of 
               NONE => data_st_Term
             | SOME(_, ctrl_st_Typ, _) => 
                 fst (mk_pair_Const sign ctrl_st_Typ data_st_Typ)
                 $ proj (the ctrl_st_exprs) $ data_st_Term
	  end);
      val s_Term  = mk_state fst (Inl []);
      val s'_Term = mk_state snd post_v;

      (* construct transition *)
      val cmd_Typ = case cmd_Typ_default_opt of SOME (t,_) => t | NONE => 
                                                                  HOLogic.unitT;
      val cmd     = case cmd_opt of SOME t => t | NONE => 
                   (case cmd_Typ_default_opt of 
                 NONE => HOLogic.unit  
               | SOME (_, SOME default_cmd) => default_cmd
               | _ => ism_error ("ISM has commands without default command " ^ 
                    "but subsection \"cmd\" is missing in rule " ^ quote name));

      val (is_s_Term, is_s_Typ) = 
           mk_pair_Term_Typ sign (is_Term, msgss_Typ) (s_Term, st_Typ);
      val (os_s'_Term, os_s'_Typ) = 
           mk_pair_Term_Typ sign (os_Term, msgss_Typ) (s'_Term, st_Typ);
      val (trans_Term, trans_Typ) = mk_pair_Term_Typ sign (is_s_Term, is_s_Typ) 
        (mk_pair_Term_Typ sign (cmd, cmd_Typ) (os_s'_Term, os_s'_Typ));

      (* construct conclusion *)
      val elem_Term = Const("op :", 
                    trans_Typ --> HOLogic.mk_setT trans_Typ --> HOLogic.boolT);
      val concl_Term = HOLogic.mk_Trueprop 
                       (elem_Term $ trans_Term $ param_transs_Term);

      (* construct rule with premissions *)
      val rule_Term = mk_prems sign concl_Term (rev pres)

      (* introduction rule *)
      val args' = map (Attrib.global_attribute thy) args;
      val intr = ((name, rule_Term), args');
    in
      intr
    end;


(** add transition set **)
(* If you use something newer than Isabelle2005, use OldInductivePackage instead *)
local
  open InductivePackage
in
fun add_transs prep_rule types param_transs_Term transs_Typ [] (thy, thms) = 
                                                               (thy, thms)
|   add_transs prep_rule types param_transs_Term transs_Typ transs (thy, thms) =
    let
      val sign = Theory.sign_of thy;
      val intros = map (mk_intros prep_rule thy types param_transs_Term) transs;
      val (thy', thm) = thy
        |> add_inductive_i true false "" false false false 
                                            [param_transs_Term] intros [];
      val thms' = thm :: thms;

      val sign' = Theory.sign_of thy';
      val elim_Term  = HOLogic.mk_Trueprop (HOLogic.mk_mem 
                 (Free ("t", HOLogic.dest_setT transs_Typ), param_transs_Term));
      val thy'' = inductive_cases_i [(("elims",[]), [elim_Term])] thy';
    in
      (thy'', thms')
    end;
end;



(*** add_ism ***)

fun gen_add_ism prep_type prep_term prep_rule name params' 
         pt_type pt_set_I pt_set_O msg_type cmd_type_default_opt 
         g_st_type_opt ctrl_st data_st transs thy =
let
  val _ = message ("Constructing ism " ^ quote name ^  " ...");
  val sign = Theory.sign_of thy;

  (* open own namespace *)
  val thy_old = thy;
  val thy = Theory.add_path name thy;
  val sign = Theory.sign_of thy;

  val params = map (fn (n,ty) => (n, mk_Typ prep_type sign ty)) params';

  (* port types and terms *)
  val pt_Typ = mk_Typ prep_type sign pt_type;
  val pt_set_Typ = HOLogic.mk_setT pt_Typ;
  val pt_set_I_Term = mk_Term prep_term sign (pt_set_I, pt_set_Typ);
  val pt_set_O_Term = mk_Term prep_term sign (pt_set_O, pt_set_Typ);

  (* message type *)
  val msg_Typ = mk_Typ prep_type sign msg_type;
  
  (* command type *)
  val cmd_Typ_default_opt = option_map (fn (cmd_type, default) => 
                              let val cmd_Typ = mk_Typ prep_type sign cmd_type 
                              in (cmd_Typ,
                                  option_map (fn t =>
                                  mk_Term prep_term sign (t, cmd_Typ)) default)
                              end) cmd_type_default_opt;
  val cmd_Typ = case cmd_Typ_default_opt of NONE       => HOLogic.unitT 
                                          | SOME (t,_) => t;

  (* states *)
  (* global state type *)
  val g_st_Typ_opt = option_map (mk_Typ prep_type sign) g_st_type_opt;
  (* local state initial terms; type*)
  val ctrl_st_Typ_Term_opt = mk_state_opt prep_type prep_term sign ctrl_st;
  val data_st_Typ_Term_opt = mk_state_opt prep_type prep_term sign data_st;
  val l_st_Typ =
    case ctrl_st_Typ_Term_opt of NONE =>
          (case data_st_Typ_Term_opt of NONE =>
                let val _ = ism_error (no_state_error name) in dummyT end
            | SOME(_, data_st_Typ, _) => data_st_Typ)
    | SOME(_, ctrl_st_Typ, ctrl_st_var_dummy) =>
      case data_st_Typ_Term_opt of NONE => ctrl_st_Typ
        | SOME(_, data_st_Typ, data_st_var_dummy) => 
                 (ctrl_st_var_dummy = data_st_var_dummy; (* forcing equal type*)
                  HOLogic.mk_prodT(ctrl_st_Typ, data_st_Typ));

  (* actual state type *)
  val st_Typ = case g_st_Typ_opt of NONE           => l_st_Typ 
                                  | SOME(g_st_Typ) => g_st_Typ;

  (* ism type *)
  val ism_Typ = Type (Sign.intern_type sign "ism", 
                      [cmd_Typ, pt_Typ, msg_Typ, st_Typ]);
  val param_ism_Typ = Library.foldr (op -->) (map snd params, ism_Typ);
  val ism_name = "ism";

  (* transitions *)
  val trans_Typ = Type (Sign.intern_type sign "trans", 
                        [cmd_Typ, pt_Typ, msg_Typ, st_Typ]);
  val transs_Typ = HOLogic.mk_setT trans_Typ;
  val param_transs_Typ = Library.foldr (op -->) (map snd params, transs_Typ);
  val transs_name = "transs";

  (* add transs and ism constants *)
  val thy = thy
    |> Theory.add_consts_i [(transs_name, param_transs_Typ, NoSyn)]
    |> Theory.add_consts_i [(ism_name, param_ism_Typ, NoSyn)]
  val sign = Theory.sign_of thy

  (* add transs (inductive definition) *)
  val transs_Term = Sign.intern_term sign 
                    (term_of (read_cterm sign (transs_name, param_transs_Typ)));
  val param_transs_Term = Library.foldl (op $) (transs_Term, map Free params);
  val types = (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ, 
               ctrl_st_Typ_Term_opt,data_st_Typ_Term_opt,l_st_Typ,g_st_Typ_opt);
  val (thy, thms) = add_transs prep_rule types param_transs_Term transs_Typ 
                    transs (thy, [])
  
  (* add syntax (symbols) and translations *)
  val sign = Theory.sign_of thy;
  val syn_name = name ^ "_" ^ transs_name ^ "_syntax";
  val msgss_Typ = pt_Typ --> mk_listT msg_Typ;
  val symbols_Typ = Library.foldr (op -->) (map snd params, 
                       msgss_Typ --> msgss_Typ --> msgss_Typ --> 
                       st_Typ --> st_Typ --> HOLogic.boolT);
  val mixfix_String = (escape_underscores name) ^ Library.foldr (op ^)
    (map (fn _ => ":_") params, 
    ":\<langle>_,_\<rangle> \<midarrow>_\<rightarrow>" ^
    " \<langle>_,_\<rangle>");
  val mixfix = Mixfix (mixfix_String, Library.foldr (op ::) 
                   (map (fn _ => 10) params, [10, 10, 10, 10, 10]), 60);
  val param_vars = map (Variable o fst) params;
  fun mk_appl cname = Syntax.mk_appl (Constant cname);
  val trrule = mk_appl syn_name (param_vars @ 
                                 map Variable ["p", "s", "c", "p'", "s'"]);
  datatype name_variant = Short | Medium | Long;
  fun nv n Short  =              NameSpace.base n
  |   nv n Medium = name ^ "." ^ NameSpace.base n
  |   nv n Long  = n;
  fun state_constructor s v = case g_st_Typ_opt of NONE => s
    | SOME g_st_Typ => mk_appl (nv (fetch_constr l_st_Typ g_st_Typ thy) v) [s];
  fun tuple_AST v = mk_appl "_tuple" 
      [mk_appl "_tuple"      [                   Variable "p", 
       mk_appl "_tuple_arg"  [state_constructor (Variable "s") v]],
       mk_appl "_tuple_args" [                   Variable "c", 
       mk_appl "_tuple_args" [                   Variable "p'", 
       mk_appl "_tuple_arg"  [state_constructor (Variable "s'") v]]]];
  fun trrule' (v1, v2) = mk_appl "op :" [tuple_AST v1,
            mk_appl (nv (Sign.intern_const sign transs_name) v2) param_vars]; 
  val thy = thy
    |> Theory.add_modesyntax_i ("symbols", true) 
                               [(syn_name, symbols_Typ, mixfix)] 
    |> Theory.add_trrules_i 
        (Syntax.ParsePrintRule (trrule, trrule' (Long, Long))::
         (* additionally for robustness wrt short/long constant names: *)
         map (fn vv => Syntax.PrintRule (trrule, trrule' vv))
             [(Long, Medium), (Long, Short), (Medium, Long), (Medium, Medium), 
              (Medium, Short), (Short, Long), (Short, Medium), (Short, Short)]);

  (* initialize ism record *)
  val sign = Theory.sign_of thy;
  val typname = Sign.intern_type sign;
  val constname = Sign.intern_const sign;

  val field_Typ5 = HOLogic.unitT;
  val field_Term5 = HOLogic.unit;

  (* initialize trans field *)
  val field_Typ4 = Type (typname "trans_field_type", [transs_Typ, field_Typ5]);
  val field_Term4 = Const (constname "trans_field", 
                           transs_Typ --> field_Typ5 --> field_Typ4) 
                    $ param_transs_Term $ field_Term5;

  (* initialize init field *)
  val field_Typ3 = Type (typname "init_ext_type", [st_Typ, field_Typ4]);
  val field_Term3 =
    let
      val init_Term = 
	case ctrl_st_Typ_Term_opt of NONE =>
            (case data_st_Typ_Term_opt of NONE => dummy_Term
              | SOME(data_st_Term, _, _) => data_st_Term)
      | SOME(ctrl_st_Term, ctrl_st_Typ, _) =>
        case data_st_Typ_Term_opt of NONE => ctrl_st_Term
          | SOME(data_st_Term, data_st_Typ, _) =>
            let
	      val (pair_Const, _) = mk_pair_Const sign ctrl_st_Typ data_st_Typ;
            in
              pair_Const $ ctrl_st_Term $ data_st_Term
            end;
      val init_Term' = mk_global_st_term l_st_Typ g_st_Typ_opt thy init_Term;
    in
      Const (constname "init_field", st_Typ --> field_Typ4 --> field_Typ3)
      $ init_Term' $ field_Term4
    end;

  (* initialize outputs field *)
  val field_Typ2 = Type (typname "outputs_ext_type", [pt_set_Typ,field_Typ3]);
  val field_Term2 = Const (constname "outputs_field", 
                           pt_set_Typ --> field_Typ3 --> field_Typ2) 
                    $ pt_set_O_Term $ field_Term3;

  (* initialize inputs field *)
  val field_Typ1 = Type (typname "inputs_ext_type", [pt_set_Typ,field_Typ2]);
  val field_Term1 = Const (constname "inputs_field", 
                           pt_set_Typ --> field_Typ2 --> field_Typ1) 
                    $ pt_set_I_Term $ field_Term2;

  val def_eq = Const("==", field_Typ1 --> field_Typ1 --> propT);
  val ism_Term = Library.foldl (op $) (Const (constname ism_name, param_ism_Typ), 
                               map Free params);
  val ism_Term = Library.foldl (op $) (Const (constname ism_name, param_ism_Typ), map Free params);
  val tqez = def_eq $ ism_Term $ field_Term1;

  val tqueque = cterm_of sign tqez;

  val ism_def_Term = term_of tqueque;

  val (thy, thm) = thy
    |> PureThy.add_defs_i false [(("ism_def", ism_def_Term), [])];

  (* close own namespace *)
(* FIXME: Ich kann nicht behaupten dass ich diese Aenderung verstehe -- Thomas *)
(*  val thy = Theory.parent_path (Thm.theory_of_thm thy); *)
   val thy = thy_old;
in
  (thy, thms, thm)
end;


fun assoc_string ([], (key:string)) = NONE
  | assoc_string ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_string (pairs, key);

fun read_rule sign sTs = 
  ((fst o (read_def_cterms (sign, K NONE, 
        (fn (x, ~1) => (curry assoc_string o typ_tfrees) TypeInfer.logicT x))
    (((Library.flat o (map ((map fst) o typ_tfrees))) o map snd) sTs) true)) sTs);
fun cterm_of' sign = cterm_of sign o fst;
fun rule_of sign = map (cterm_of' sign);

val add_ism   = gen_add_ism read_ctyp read_cterm read_rule;
val add_ism_i = gen_add_ism ctyp_of   cterm_of'  rule_of;



(*** package setup ***)

(** outer syntax **)

local structure P = OuterParse and K = OuterKeyword in

val ism_decl = 
  let val optional_init = Scan.optional (P.$$$ "init" |-- P.term) "arbitrary"
      val param = P.$$$ "(" |-- P.name -- (P.$$$ "::" |-- P.typ) --| P.$$$ ")"
      val pattern = ((                     P.term >> (fn x => (x,false)))
                     || (P.$$$ "multi" |-- P.term >> (fn x => (x,true))))
                    -- P.term 
      val state_update = P.list1 (P.name -- (P.$$$ ":=" |-- P.term)) >> Inl 
                         ||                                 P.term   >> Inr
  in (P.name -- Scan.repeat param) -- (P.$$$ "=" |--
    (P.$$$ "ports" |-- P.typ --
      (P.$$$ "inputs"  |-- P.term --
      (P.$$$ "outputs" |-- P.term --
    (P.$$$ "messages" |-- P.typ --
    (Scan.option (P.$$$ "commands" |-- P.typ -- 
                  Scan.option (P.$$$ "default" |-- P.typ)) --
    (P.$$$ "states" |-- Scan.option P.typ --
      (Scan.option (P.$$$ "control" |-- P.typ -- 
              (optional_init >> (fn x => (x, "dummy_control_var")))) --
      (Scan.option (P.$$$ "data"    |-- P.typ -- 
              (optional_init --
                (Scan.optional (P.$$$ "name" |-- P.term) "s"))) --
    Scan.optional (P.$$$ "transitions" |--
      Scan.repeat1 (P.thm_name ":" -- 
        (Scan.option (P.term -- ((P.$$$ "\<rightarrow>" || P.$$$ "->") |--  
                      P.term)) --
        (Scan.optional (P.$$$ "pre"  |--  P.list1 P.term ) [] --
        (Scan.optional (P.$$$ "in"   |--  P.list1 pattern) [] --
        (Scan.optional (P.$$$ "out"  |--  P.list1 pattern) [] --
        (Scan.option   (P.$$$ "cmd"  |--          P.term)     --
        (Scan.optional (P.$$$ "post" |--  state_update) (Inl []))
       ))))))) []
    ))))))))) end;

val ismP =
  OuterSyntax.command "ism" "define Interacting State Machine" K.thy_decl
   (ism_decl >> (fn ((name, params), (pt_type, (pt_set_I, (pt_set_O, (msg_type, 
   (cmd_type_default_opt, (g_st_type_opt, (ctrl_st, (data_st, transs))))))))) =>
      Toplevel.theory
        (#1 o add_ism name params pt_type pt_set_I pt_set_O msg_type 
                   cmd_type_default_opt g_st_type_opt ctrl_st data_st transs)));

val _ = OuterSyntax.add_keywords 
["ports", "inputs", "outputs", "messages", "commands", "default",
 "states", "control", "init", "data", "name", "multi",
 "transitions", "\<rightarrow>", "->", "pre", "in", "out", "cmd", "post",":="];

val _ = OuterSyntax.add_parsers [ismP];

end; (* local *)


end;
*}
end
(*>*)

