(*  Title:      ISM/SLE66.thy
    Id:         $Id: SLE66.thy,v 1.33 2006/03/16 18:03:30 dvo Exp $
    Author:     David von Oheimb
    Copyright:  (C) 2002 Siemens AG

*)

header {* An improvement of the LKW Model for the Infineon SLE 66 *}

theory SLE66 imports ISM_package  --{* we build on the general ISM definitions *}
begin
(*<*)
ML {* Pretty.setmargin 169; quick_and_dirty:=true *}
(*>*)
text {* First we have to define a number of entities (types, logical constants,
etc.) acting as building blocks for the actual ISM theory section.
In order to keep the model as abstract as possible, which makes it less bulky
to read and simplifies the proofs, we often use underspecification. This
important modeling technique means that for part of the types and constants
we do not give full definitions but only declarations of their names. 
We even do not make the encryption of data in the memory components explicit.
*}

subsection "Names"

text {* 
Objects stored on the chip may be either functions or data and are referred to 
by object names. The type of these names, @{term on}, is the disjoint sum of
function names @{term fn} and data object names @{term dn}, which are not 
further specified:
*}
typedecl fn               --{* function name *}
typedecl dn               --{* data object name *}
datatype on = F fn | D dn --{* object name *}

text {* 
Objects are classified as security-relevant (demanding secrecy and integrity)
by including their names in the sets @{term F_Sec} or @{term D_Sec}, whose 
disjoint union is called @{term Sec}.
In order to meet the additional requirements of \cite{SCPP01}, the domain of 
security relevant functions @{term F_Sec} of the original LKW 
model has been refined to the disjoint union of @{term F_PSec}
and @{term F_ASec}, which control the protection of the 
processor and application functionality, respectively.

In the following theory sections, we declare a list of constants together with
their types. We define only part of them, and for part of the remaining ones
we give the essential properties in the form of axioms:
*}
consts 
  f_SN   :: "fn"      --{* the name of the function giving the serial number *}
  FTest0 :: "fn set"  --{* the names of test functions of phase 0 *}
  FTest1 :: "fn set"  --{* the names of test functions of phase 1 *}
  FTest  :: "fn set"  --{* the names of all test functions *}
  F_Sec  :: "fn set"  --{* the names of all security-relevant functions *}
  F_PSec :: "fn set"  --{* the subset of @{term F_Sec} relevant for the processor *}
  F_ASec :: "fn set"  --{* the names of @{term F_Sec} relevant for applications *}
  F_NSec :: "fn set"  --{* the names of all non-security-relevant functions *}
  D_Sec  :: "dn set"  --{* the names of all security-relevant data objects *}
  D_PSec :: "dn set"  --{* the subset of @{term D_Sec} relevant for the processor *}
  D_ASec :: "dn set"  --{* the names of @{term D_Sec} relevant for applications *}
  D_NSec :: "dn set"  --{* the names of all non-security-relevant data objects *}
    Sec  :: "on set"  --{* the names of all security-relevant objects *}
defs
  FTest_def:  "FTest  \<equiv> FTest0 \<union> FTest1"
  F_ASec_def: "F_ASec \<equiv> F_Sec - F_PSec"
  D_ASec_def: "D_ASec \<equiv> D_Sec - D_PSec"
  F_NSec_def: "F_NSec \<equiv> -F_Sec"
  D_NSec_def: "D_NSec \<equiv> -D_Sec"
  Sec_def: "Sec \<equiv> {F fn |fn. fn \<in> F_Sec} \<union> {D dn |dn. dn \<in> D_Sec}"
axioms 
  FTest01_disjunct: "FTest0 \<inter> FTest1 = {}"
  f_SN_not_FTest:   "f_SN \<notin> FTest"
  F_PSec_is_Sec:    "F_PSec \<subseteq> F_Sec"
  FTest_is_PSec:    "FTest \<subseteq> F_PSec"
(*<*)
lemma FTest01_disjunctE: "\<lbrakk>f \<in> FTest0; f \<in> FTest1\<rbrakk> \<Longrightarrow> R"
by (cut_tac FTest01_disjunct, fast)

lemmas f_SN_not_FTest [simp]
lemma [simp]: "(- FTest0 \<inter> - FTest) = (- FTest)" by (auto simp: FTest_def)
lemma [simp]: "(- FTest1 \<inter> - FTest) = (- FTest)" by (auto simp: FTest_def)

lemma F_Sec_in_Sec: "fn \<in> F_Sec \<Longrightarrow>  F fn \<in> Sec"
by (auto simp add: Sec_def)

lemma FTest0_is_FTest: "f \<in> FTest0 \<Longrightarrow> f \<in> FTest"
by (auto simp: FTest_def)

lemma FTest1_is_FTest: "f \<in> FTest1 \<Longrightarrow> f \<in> FTest"
by (auto simp: FTest_def)

lemma FTestD: "f \<in> FTest \<Longrightarrow> f \<in> FTest0 \<or> f \<in> FTest1"
by (auto simp: FTest_def)

lemma FTest_is_Sec: "f \<in> FTest \<Longrightarrow> f \<in> F_Sec"
by (cut_tac FTest_is_PSec F_PSec_is_Sec, fast)

lemma FTest_not_ASec: "\<lbrakk>f \<in> FTest; f \<in> F_ASec\<rbrakk> \<Longrightarrow> R"
by (cut_tac FTest_is_PSec, unfold F_ASec_def, fast)

(*>*)
text {* \vspace{-6ex} *}
subsection "State"

text {* 
The abstract state of an SLE 66 chip is a pair, where the first component is the
phase in the processor life cycle:
\vspace{-.5ex}
*}
datatype ph = P0 | P1 | P2 | Error
(*<*)
(* unused *)consts rank :: "ph \<Rightarrow> nat"
(* unused *)primrec
  "rank P0 = 0"
  "rank P1 = 1"
  "rank P2 = 2"
  "rank Error = 3"
lemma rank_bound [simp]: "rank ph \<le> 3"
apply (induct ph)
apply auto
done

(*>*)
text {*
\vspace{-1ex}
We introduce the type @{term val} for any values, i.e.\ function code or data 
stored or processed by the chip. The only thing we need to know about the type
@{term val} is that the serial number of the chip belongs to it.
\vspace{-1ex}
*}
typedecl val --{* data and function values *}
consts SN :: val --{* serial number *}
text {*
The second state component is a record of two partial functions, @{term valF} 
and @{term valD}, mapping function and data object names to values:
\vspace{-.5ex}
*}
record chip_data = 
  valF :: "fn \<rightharpoonup> val"
  valD :: "dn \<rightharpoonup> val"
(*<*)
translations "chip_data" \<leftharpoondown> (type) "\<lparr>valF :: fn \<rightharpoonup> val, valD :: dn \<rightharpoonup> val\<rparr>"
(*>*)
text {*
\vspace{-1.5ex}
The function @{term val} takes an argument of type @{term chip_data} and yields 
a partial function lifting @{term valF} and @{term valD} to general object
 names of type @{typ on}:
\vspace{-1ex}
*}
constdefs 
  val :: "chip_data \<Rightarrow> on \<rightharpoonup> val"
 "val s on \<equiv> case on of F fn \<Rightarrow> valF s fn  | D dn \<Rightarrow> valD s dn"
(*<*)
lemma valF2val: "valF s fn = Some v \<Longrightarrow> val s (F fn) = Some v"
by (auto simp add: val_def)

(*>*)
text {*
\paper{\pagebreak}
Having defined the two components of the processor state, we can now give the
definition of the overall state:
*}
types SLE66_state = "ph \<times> chip_data"
(*<*)
translations "SLE66_state" \<leftharpoondown> (type) "ph \<times> chip_data"
(*>*)
text {*
We will often need to refer to the set of functions available in the current
state, therefore we introduce an auxiliary function @{term fct} that yields the
domain of @{term valF}:
*}
constdefs
  fct :: "chip_data \<Rightarrow> fn set"
 "fct s \<equiv> dom (valF s)"
(*<*)
lemma dom_valF_fct [simp]: "f \<in> dom (valF s) = (f \<in> fct s)"
by (auto simp: fct_def)

lemma fct_upd_valF [simp]: "fct (s\<lparr>valF := f\<rparr>) = dom f" 
by(auto simp: fct_def)

(*>*)
text {*
\vspace{-1ex}
We declare three further auxiliary functions that denote the results and state
changes of a processor function (including test functions):
*}
consts
 "output"   :: "fn \<Rightarrow> chip_data \<Rightarrow> val"
 "change"   :: "fn \<Rightarrow> chip_data \<Rightarrow> chip_data"  
                              --{* @{term change} is unused for test functions *}
 "positive" :: "val \<Rightarrow> bool"  --{* check for positive test outcome *}

subsection "Further ISM section ingredients"

text {*
We need only two port names, one for input to the chip and one for its output:
*}
datatype interface = In | Out

text {* 
SLE 66 commands provide information on the subjects issuing them.
There is a special subject @{term Pmf} denoting the processor manufacturer.
*}
typedecl sb      (* --{* subjects *}*)
consts Pmf :: sb (* --{* processor manufacturer *}*)

text {*
Possible input consists of either the two kinds of SLE 66 commands modeling 
function execution and function code loading operations or the @{text Spy} 
operation, which models attacks that may reveal information stored on the chip
and may corrupt the chip memories. Output of the SLE 66 may be the result value
of a (regular) function or an indication of success or failure.
*}
datatype message = 
  Exec sb fn | Load sb fn val | Spy on  --{* input *}
| Val val | Ok | No                     --{* output *}

text {*
The subjects performing regular commands identify themselves to the 
chip via physical means. The actual authentication mechanism, as well as many 
other implementation details, is confidential and beyond the scope of this 
article anyway. Here we just declare an auxiliary function that yields the 
subject issuing a (regular) command:
*}
consts subject :: "message \<Rightarrow> sb"
primrec
  "subject (Exec sb fn  ) = sb"
  "subject (Load sb fn v) = sb"
text {*
\paper{\pagebreak}
*}
subsection "ISM definition"

text {*
Having defined its various parameters, we can finally give the theory section 
that specifies the SLE 66 model as an ISM: \label{SLE66-trans}
*}
ism SLE66 =
  ports interface
    inputs \<spacespace>"{In}"
    outputs "{Out}"
  messages message
  states
    control ph init "P0" 
    data   chip_data name "s" -- {* The data state variable is called @{term s}.
Note that the initial data state is left unspecified and thus is arbitrary, 
which is a good example of underspecification since its actual value is 
immaterial for the security properties we are interested in.
\vspace{1ex}
*}
  transitions

--{*
Rule \texttt{R00} specifies execution of a test function @{term f} from
the set @{term FTest0}
by the processor manufacturer @{term Pmf} in the initial phase @{term P0}.
If the test is successful then the SLE 66 enters the next phase @{term P1}, 
answers with @{term Ok}, and disables the test functions @{term FTest0}.
As specified by the \texttt{data} theory subsection just above, the variable
@{term s} denotes the current data state of the ISM at the beginning of the
transition. Thus, for example, @{term "fct s"} means the functions currently
available. The operator `@{text "\<lfloor>"}' below restricts a partial function, 
in this case @{term "valF s"}, to the given set, in this case the complement 
of @{term FTest0}.

Rule \texttt{R00} is typical for 
interactions of the SLE 66 in the sense that a single input triggers a single
output. Note that the direct relation of input and output is expressed easily
using ISMs, whereas using IOAs, two transitions would be required whose
relation would be cumbersome to express and to use during verification.
 *}
  R00: P0 \<rightarrow> P1
    pre "f \<in> fct s\<inter>FTest0", "positive (output f s)"
    in  In  "[Exec Pmf f]"
    out Out "[Ok]"
    post valF := "valF s\<lfloor>(-FTest0)"
--{*
Rule \texttt{R01} is analogous to \texttt{R00}, but specifies that if the test 
function @{term f} is from @{term FTest1} rather than @{term FTest0} then phase 
@{term P1} is skipped and the chip enters @{term P2} immediately, disabling all 
test functions @{term FTest}:
 *}
  R01: P0 \<rightarrow> P2
    pre "f \<in> fct s\<inter>FTest1", "positive (output f s)"
    in  In  "[Exec Pmf f]"
    out Out "[Ok]"
    post valF := "valF s\<lfloor>(-FTest)"
--{*
If in @{term P0} a test function gives a negative result then the
chip enters the @{term Error} phase and the output is @{term No}:
 *}
  R02: P0 \<rightarrow> Error
    pre "f \<in> fct s\<inter>FTest0", "\<not>positive (output f s)"
    in  In  "[Exec Pmf f]"
    out Out "[No]"
--{*
Any other function call issued by the processor manufacturer in @{term P0} has
the standard consequences: The function result is output and the data state
changed (according to the semantics of the function which is not further
specfied). Note that by the form of postcondition used, the whole data
state (consisting of @{term valF} and @{term valD} here) is replaced by the
given value: the denotation of @{term "change f s"}.
 *}
  R03: P0 \<rightarrow> P0
    pre "f \<in> fct s - FTest"
    in  In  "[Exec Pmf f]"
    out Out "[Val (output f s)]"
    post "change f s"
--{*
In all remaining cases for phase 0, the attempted function execution is ignored 
and the output is @{term No}:
 *}
  R04: P0 \<rightarrow> P0
    pre "sb \<noteq> Pmf \<or> f \<notin> fct s"
    in  In  "[Exec sb f]"
    out Out "[No]"
--{*
This ends the specifications of transitions originating in @{term P0}.
\vspace{2ex}

The specifications of transitions originating in @{term P1} are fully
analogous to the rules \texttt{R00}, \texttt{R02}, \texttt{R03}, and
\texttt{R04}, just replacing @{term P0} by @{term P1}, @{term P1} by @{term P2},
and @{term FTest0} by @{term FTest1}: \hspace*{-2ex} 
*}
  R11: P1 \<rightarrow> P2
    pre "f \<in> fct s\<inter>FTest1", "positive (output f s)"
    in  In  "[Exec Pmf f]"
    out Out "[Ok]"
    post valF := "valF s\<lfloor>(-FTest1)"
  R12: P1 \<rightarrow> Error 
    pre "f \<in> fct s\<inter>FTest1", "\<not>positive (output f s)"
    in  In  "[Exec Pmf f]"
    out Out "[No]"
  R13: P1 \<rightarrow> P1
    pre "f \<in> fct s - FTest1"
    in  In  "[Exec Pmf f]"
    out Out "[Val (output f s)]"
    post "change f s"
  R14: P1 \<rightarrow> P1
    pre "sb \<noteq> Pmf \<or> f \<notin> fct s"
    in  In  "[Exec sb f]"
    out Out "[No]"

--{*
The rules \texttt{R21} and \texttt{R22} specify function calls in @{term P2}
analogously to \texttt{R03} and \texttt{R04}, except that any subject is
allowed to issue them:
*}
  R21: P2 \<rightarrow> P2
    pre "f \<in> fct s"
    in  In  "[Exec sb f]"
    out Out "[Val (output f s)]"
    post "change f s"
  R22: P2 \<rightarrow> P2
    pre "f \<notin> fct s"
    in  In  "[Exec sb f]"
    out Out "[No]"

--{*
In the @{term Error} phase, the only function that may be called is chip
identification, yielding the serial number @{term SN}. All other cases yield 
@{term No}:
*}
  R31: Error \<rightarrow> Error
    pre "f_SN \<in> fct s"
    in  In  "[Exec sb f_SN]"
    out Out "[Val SN]"
  R32: Error \<rightarrow> Error
    pre "f \<notin> fct s\<inter>{f_SN}"
    in  In  "[Exec sb f]"
    out Out "[No]"

--{*
The rules \texttt{R41} and \texttt{R42} specify the behavior of the 
@{term Load} operation, which is allowed only for the processor manufacturer
and only in the upload phase @{term P1}. If allowed, @{term valF} is updated 
at the position @{term f} with the new function value @{term v}.

In contrast to the original LKW model \cite{LKW-IE00}, the @{term Load} 
operation may upload not only non-security-relevant functions but also 
functions of the application security domain (as long as no such function 
of the same name is already present). \hspace*{-2ex}
*}
  R41: P1 \<rightarrow> P1
    pre "f \<in> F_NSec \<union> (F_ASec - fct s)"
    in  In  "[Load Pmf f v]"
    out Out "[Ok]"
    post valF := "valF s(f\<mapsto>v)"
  R42: ph \<rightarrow> ph
    pre "f \<notin> F_NSec \<union> (F_ASec - fct s) \<or> sb \<noteq> Pmf \<or> ph \<noteq> P1"
    in  In  "[Load sb f v]"
    out Out "[No]"
--{*
Note that the rule \texttt{R42} is generic in the sense that it applies to more
than one control state of the ISM, namely all phases except @{term P1}.
*}

--{*
The rules \texttt{R51} thru \texttt{R53} specify the possible reactions of the
chip to attacks, modeled by the @{term Spy} operation. If the attacker attempts
to read a non-secret object whose name is @{term on} and the chip is not in the 
@{term Error} phase, the access may be granted, yielding the desired value 
(if any):
*}
  R51: ph \<rightarrow> ph
    pre "on \<notin> Sec", "ph \<noteq> Error"
    in  In  "[Spy on]"
    out Out "case val s on of None \<Rightarrow> [] | Some v \<Rightarrow> [Val v]"
--{*
Rule \texttt{R52} specifies the typical reaction of the SLE 66 upon attacks
trying to read a secret object while tampering with the chip: it may be
unable to prevent that the desired value is output, but in any case it reaches
the @{term Error} phase from which no further secrets may be obtained,
as specified by the rules \texttt{R31}, \texttt{R32}, and \texttt{R53}.
*}
  R52: ph \<rightarrow> Error
    pre "on \<in> Sec", "v \<in> {[],[Val (the (val s on))]}", "ph \<noteq> Error"
    in  In  "[Spy on]"
    out Out "v"
    post "any"
--{*
Note that \texttt{R52} describes two sorts of nondeterminism: @{term v} denotes 
either the empty output or the singleton output giving the desired value, and 
the attack may corrupt the function and data stores arbitrarily.

\paper{\pagebreak}
There are also cases where the chip can resist an attack without any damage
and without any leakage of secrets, such that there is no need to enter
the @{term Error} phase:
*}
  R52':ph \<rightarrow> ph
    pre "on \<in> Sec", "ph \<noteq> Error"
    in  In  "[Spy on]"
    out Out "[]"
--{*
If the chip is already in the @{term Error} phase, no further secrets can be 
obtained. The chip state may be corrupted further, but it makes sure that it
stays locked in the @{term Error} phase:
*}
  R53: Error \<rightarrow> Error
    in  In  "[Spy on]"
    out Out "[]"
    post "any"

text {* 
As expressed by the rules \texttt{R52} and \texttt{R53}, the attacker 
may obtain (the representation of) at most one secret object from the chip 
memory. 
It is interesting to observe that the leakage of one item is harmless because 
all data stored on the chip is encrypted. There are two cases to consider: 
\begin{itemize}
\item The secret obtained is the de-/encryption key itself, which is not 
  helpful to the attacker because no further data item, in particular none
  encrypted with the key, can be obtained.
\item The secret obtained is an encrypted value, which is not helpful because 
 the attacker cannot any more obtain the decryption key.
\end{itemize}
Obviously, sophisticated techniques are required to implement the specified 
reaction to physical attacks modeled by the @{term Spy} operation.
*}

subsection "ISM runs"

text {* The SLE 66 ISM just defined models the static interface of the chip as
well as all possible single state transitions that it can perform. In order to 
describe the overall behavior of the chip during its life-cycle, we can refer 
to the notions that our Isabelle implementation provides for ISMs in general:
*}
types 
  SLE66_trans = "(unit, interface, message, SLE66_state) trans"
(*<*)
translations 
  "SLE66_trans" \<leftharpoondown> (type) "(unit, interface, message, SLE66_state) trans"
(*>*)

constdefs
  Trans :: "SLE66_trans set"         --{* the set of all possible transitions *}
 "Trans \<equiv> trans SLE66.ism"

  TRuns :: "(SLE66_trans list) set"  --{* all possible transition sequences *}
 "TRuns \<equiv> truns SLE66.ism"

  Runs :: "(SLE66_state list) set"   --{* all possible sequences of states *}
 "Runs \<equiv> runs SLE66.ism"
(*<*)
lemma trans_ism [simp]: "ism.trans ism = transs"
by (simp add: ism_def)

lemma TRunsD: "\<lbrakk>ts \<in> TRuns; ((p,s),(p',s')) \<in> set ts\<rbrakk> \<Longrightarrow>
s \<in> reach ism \<and> ((p, s), p',s') \<in> transs"
by (auto simp: TRuns_def dest: truns2trans truns_lstate_reach)

(*>*)
text {* 
This concludes the system model of the SLE 66.

\paper{\pagebreak}
\subsection{Properties}
The second part of the SLE 66 security model deals with the security
properties derivable from the system model. 
*}

subsection "Security Objectives"

text {* 
In the (confidential\footnote {quotations with permission.}) original security 
requirements specification by Infineon, the security objectives for the SLE 66 
had been stated as follows.
\begin{description}
\item[SO1.] ``The hardware must be protected against unauthorised disclosure 
of security enforcing functionality.''
\item[SO2.] ``The hardware must be protected against unauthorised modification 
of security enforcing functions.''
\item[SO3.] ``The information stored in the processor's memory components must 
be protected against unauthorised access.''
\item[SO4.] ``The information stored in the processor's memory components must 
be protected against unauthorised modification.''
\item[SO5.] ``It may not occur that test functions are executed in an 
unauthorised way.''
\end{description}
Later, an additional requirement concerning the confidentiality and integrity of 
Smartcard Embedded Software, which is not part of the security enforcing 
functionality, has been added \cite[\S 4.1]{SCPP01}.

Having formally defined the SLE 66 system model, these informal statements can 
now be expressed formally as predicates on the system behavior, describing 
unambiguously and in detail which states may be reached under which 
circumstances, which data may be modified, and which output may appear on the 
output channel.

After formalizing the security objectives, it is natural to ask if the chip
behavior, as specified in the system model, actually fulfills these 
requirements. The corresponding proofs have been conducted first using pen and
paper, as reported in \cite{LKW-IE00}. Within the ISM framework, we meanwhile
have verified these properties even mechanically using Isabelle, discovering
two major flaws that will be reported in this subsection. Below we give all the 
required auxiliary definitions, the most important lemmata, and all theorems, 
together with an abstract informal description of the machine-checked proofs.
*}

subsection "Model Assumptions"

text {*
Due to the abstract specification style where e.g.\ the semantics of parts of 
the chip functionality is not fully specified, it turns out that in order to 
prove the properties, a few general axioms that augment the system model are 
required. The first one of them asserts that security-relevant functions do not 
modify security-relevant functions:
 *}
(*<*)
axioms
(*>*)
Axiom1: "f\<in>fct s\<inter>F_Sec \<Longrightarrow> valF (change f s)\<lfloor>F_Sec = valF s\<lfloor>F_Sec"
text {*
In comparison to the version of this axiom in the original model, 
the scope of functions @{term f} has been extended from 
\hspace{-.5pt}``initially available''\hspace{-.5pt} to 
\hspace{-.5pt}``security-relevant'', 
reflecting the changes to rule \texttt{R41}.
Part of the lemmas as well as the formalized security objective \texttt{FSO21} 
change accordingly.

The second axiom is very similar, stating that also non-security-relevant 
functions do not modify security-relevant functions:
*}
(*<*)
axioms
(*>*)
Axiom2: "f\<in>fct s\<inter>F_NSec \<Longrightarrow> valF (change f s)\<lfloor>F_Sec = valF s\<lfloor>F_Sec"
text {*
In order to formalize the security objective SO1 and \texttt{Axiom3}, we define 
the set @{term "ValF_Sec r"} holding all code of security-relevant functions 
in a given run (i.e., sequence of states) @{term r}. *}
constdefs
  ValF_Sec :: "SLE66_state list  \<Rightarrow> val set"
 "ValF_Sec r \<equiv> \<Union>{ran (valF s\<lfloor>F_Sec) |ph s. (ph,s) \<in> set r}"
text {*
The third (and last) axiom introduced in the LKW model states that in phase~2,
a function cannot reveal (by intentional ``guessing'' or by accident) any 
members of @{term "ValF_Sec r"}. This rather self-evident requirement is
needed for technical reasons in the proof of SO1.
*}
(*<*)
axioms
(*>*)
Axiom3: "\<lbrakk>r\<in>Runs; (P2,s)\<in>set r; f\<in>fct s\<rbrakk>\<Longrightarrow> output f s\<notin>ValF_Sec r"
text {*
\noindent A notational remark is in order here: in Isabelle formualas, 
multiple premises are bracketed using `@{text "\<lbrakk>"}' and `@{text "\<rbrakk>"}' 
and separated using `@{text ";"}'.

When machine-checking the proofs contained in \cite{LKW-IE00} with Isabelle,
we noticed that a fourth axiom was missing that makes an implicit but important 
assumption explicit: if a function object may be referenced in two (different) 
ways and one of them declares the function to be security-relevant, the other 
has to do the same.
*}
(*<*)
axioms
(*>*)
  Axiom4: "\<lbrakk>r \<in> Runs; 
            (ph, s) \<in> set r; (ph', s') \<in> set r; 
            val s n = Some v; val s' n' = Some v; 
            n \<in> Sec\<rbrakk> \<Longrightarrow> n' \<in> Sec"
(*<*)
lemma
Axiom1': "\<lbrakk>f \<in> fct s; f \<in> F_Sec; g \<in> F_Sec\<rbrakk> \<Longrightarrow> valF (change f s) g = valF s g"
apply (drule (1) IntI [THEN Axiom1])
apply (drule spec [OF expand_fun_eq [THEN iffD1], of _ _ g])
apply simp
done

lemma
Axiom2': "\<lbrakk>f \<in> fct s; f \<in> F_NSec; g \<in> F_Sec\<rbrakk> \<Longrightarrow> valF (change f s) g = valF s g"
apply (drule (1) IntI [THEN Axiom2])
apply (drule spec [OF expand_fun_eq [THEN iffD1], of _ _ g])
apply simp
done

lemma
Axiom12': "\<lbrakk>f \<in> fct s; g \<in> F_Sec\<rbrakk> \<Longrightarrow> valF (change f s) g = valF s g"
apply (case_tac "f \<in> F_NSec")
apply (auto simp: F_NSec_def elim: Axiom1' Axiom2')
done

lemma
Axiom12'': "\<lbrakk>f \<in> fct s; g \<in> F_Sec; g \<in> fct s\<rbrakk> \<Longrightarrow> g \<in> fct (change f s)"
apply (drule (1) Axiom12')
apply (auto simp: fct_def dom_def)
done

lemma
Axiom12''': "\<lbrakk>g \<in> fct (change f s); f \<in> fct s; g \<in> F_Sec\<rbrakk> \<Longrightarrow> g \<in> fct s"
apply (tactic "rotate_tac 1 1")
apply (drule (1) Axiom12')
apply (auto simp: fct_def dom_def)
done

(*>*)
text {*
The exposure of missing critical assumptions demonstrates how important 
machine support is when conducting formal analysis.
\vspace{-2ex}
*}
subsection "Theorems"

text {*
Finally, we translate the five informal security objectives to Isabelle formulas
and prove them within the system. It is instructive to compare the formal 
versions of the security objectives FSOx below with the informal ones, SOx, 
given above.
*}
(*<*)
lemma ValF_Sec_is_Sec:
 "\<lbrakk>r \<in> Runs; (ph,s) \<in> set r; v \<in> ValF_Sec r; val s n = Some v\<rbrakk> \<Longrightarrow> n \<in> Sec"
apply (unfold ValF_Sec_def)
apply (auto dest!: ran_restrictD)
apply (drule valF2val)
apply (drule F_Sec_in_Sec)
apply (force dest!: Axiom4)
done

(* unused *)lemma 
  Lemma2: "\<lbrakk>((p,(ph,s)),c,(p',(ph',s'))) \<in> Trans\<rbrakk> \<Longrightarrow> rank ph \<le> rank ph'"
apply (auto simp: Trans_def)
apply (auto elim!: SLE66.transs.elims)
done
(*>*)

text {*
The formalization of SO1, called \texttt{FSO1}, states that in any sequence 
@{term ts}
of transitions performed by the chip, if the chip outputs any value @{term v} 
representing the code of any security-relevant function during its hitherto 
life, then the error state is entered or the output was in response to a
function execution request by the processor manufacturer:
\paper{\pagebreak}
*}
theorem FSO1: "\<lbrakk>ts \<in> TRuns; ((p,(ph,s)),c,(p',(ph',s'))) \<in> set ts;
                 p' Out = [Val v]; v \<in> ValF_Sec (truns2runs ts)\<rbrakk> \<Longrightarrow>
                ph' = Error \<or> (\<exists>fn. p In = [Exec Pmf fn])"
(*<*)
apply (unfold TRuns_def)
apply (frule (1) truns2trans)
apply (drule truns2runs)
apply (fold Runs_def)
apply (drule lstate_in_truns2runs)
apply (auto elim!: SLE66.transs.elims split: option.split_asm)
apply  (fast dest!: Axiom3)
apply (fast dest!: ValF_Sec_is_Sec)
done
(*>*)
text {*
\vspace{-1ex}
The proof of \texttt{FSO1} proceeds by unfolding some definitions, e.g.\ of 
the SLE 66 
ISM, applying properties of auxiliary concepts like @{term truns2runs}, and
a case split on all possible transitions. Isabelle can solve most of the cases 
automatically (with straightforward term rewriting and purely predicate-logical 
reasoning), except for two: the case of rule \texttt{R21} is handled using 
\texttt{Axiom3}, and for \texttt{R51} we rely on the property 
\isa{\isachardoublequote}@{thm ValF_Sec_is_Sec [no_vars]}\isa{\isachardoublequote} which in turn relies on \texttt{Axiom4}.

A more elaborate formalization of SO1 and SO3 taking into account also
indirect and partial information flow is motivated and sketched in
\cite{DvO-Noninfluence04}.

Like in the original LKW model, the translation of SO2 splits into two parts.
\texttt{FSO21'} states that for any (even unreachable) transition not ending in 
the error phase, if a security-relevant function @{term g} is present in both 
the pre-state and the post-state, the code associated with it stays the same:
*}
theorem FSO21': "\<lbrakk>((p,(ph,s)),c,(p',(ph',s')))\<in>Trans; ph' \<noteq> Error;
                g \<in> fct s\<inter>fct s'\<inter>F_Sec\<rbrakk> \<Longrightarrow> valF s' g = valF s g"
(*<*)
by (auto simp: Trans_def F_NSec_def elim!: SLE66.transs.elims Axiom12')
(*>*)
text {*
\vspace{-3ex}
This property is a generalization of the original \texttt{FSO21}, reflecting 
the extensions made to the @{term Load} operation in rule \texttt{R41}:
Here we do not compare the initial and current value of @{term g} but
the previous and current one, which takes into account also functions 
added in the meantime.

The proof of this property is --- as usual --- by 
%induction on the construction of all runs the SLE 66 ISM can perform, 
%which boils down to a 
case distinction over all possible transitions. 
Most cases are trivial except for those where 
function execution may change the stored objects, which are described by the 
rules \texttt{R03}, \texttt{R13}, and \texttt{R21}. Here an argumentation 
about the invariance of security-relevant functions @{term g} is needed, which 
follows easily from \texttt{Axiom1} and \texttt{Axiom2}.

Similarly to \texttt{FSO21'}, \texttt{FSO22} states that for any transition 
within the same phase that is not the error phase, the set of existing 
security-relevant functions is non-decreasing:
*}
theorem FSO22: "\<lbrakk>((p,(ph,s)),c,(p',(ph',s')))\<in>Trans; ph' \<noteq> Error;
                 ph = ph'\<rbrakk> \<Longrightarrow> fct s\<inter>F_Sec \<subseteq> fct s'\<inter>F_Sec"
(*<*)
by (auto simp: Trans_def elim!: SLE66.transs.elims Axiom12'')
(*>*)
text {*
\vspace{-3ex}
Not surprisingly, the proof of this property is completely analougous.

\texttt{FSO3} states that if the attacker obtains a result trying to get hold 
of a security-relevant data object @{term on}, then the chip enters (or stays 
in) the error phase:
*}
theorem FSO3:"\<lbrakk>((p,(ph,s)),c,(p',(ph',s')))\<in>Trans; p In = [Spy on];
               on \<in> Sec; p' Out \<noteq> []\<rbrakk> \<Longrightarrow> ph' = Error"
(*<*)
by (auto simp: Trans_def elim!: SLE66.transs.elims)
(*>*)
text {*
\vspace{-2ex}
The proof is done simply by case distinction.

It is easy to see and to prove that once the chip is in the error phase,
it stays there and the only possible output is the serial number:
*}
theorem FSO13:"\<lbrakk>((p,(ph,s)),c,(p',(ph',s')))\<in>Trans; ph = Error;
                p' Out = [Val v]\<rbrakk> \<Longrightarrow> ph' = Error \<and> v = SN"
(*<*)
by (auto simp: Trans_def elim!: SLE66.transs.elims)
(*>*)
text {*
\texttt{FSO4} states that any transition not ending in the error phase but 
changing the state does this in a well-behaved way: 
@{term s'} is derived from @{term s} via the desired effect of executing 
an existing function, or there is a phase change where only the test functions
may be affected, or only a single function @{term f} is affected by a
@{term Load} operation:
\paper{\pagebreak}
*}
theorem FSO4: 
"\<lbrakk>((p,(ph,s)),c,(p',(ph',s'))) \<in> Trans; ph' \<noteq> Error\<rbrakk> \<Longrightarrow> 
 s' = s \<or>
 (\<exists>sb f  . p In = [Exec sb f  ] \<and> f \<in> fct s \<and> s' = change f s) \<or>
 (ph'\<noteq>ph \<and> valD s' = valD s \<and> valF s'\<lfloor>(-FTest) = valF s\<lfloor>(-FTest))\<or>
 (\<exists>sb f v. p In = [Load sb f v] \<and> 
            valD s' = valD s \<and> valF s'\<lfloor>(-{f}  ) = valF s\<lfloor>(-{f}  ))"
(*<*)
by (auto simp: Trans_def elim!: SLE66.transs.elims)
(*>*)
text {*
\vspace{-3ex}
The proof is also straightforward by case distinction.

A second omission of the LKW model was that in the proof of the security 
objective \texttt{FSO5} an argumentation about the accessibility of certain 
functions was
not given in a rigorous way. We fix this by introducing an auxiliary property
 (where, as typical with invariants, finding the appropriate one is the main 
challenge) and proving it to be an invariant of the ISM. The invariant states
that in phase 1, the test functions from @{term FTest0} have been disabled,
and in phase 2, all test functions have been disabled:
*}
constdefs
  no_FTest_invariant :: "SLE66_state \<Rightarrow> bool"
 "no_FTest_invariant \<equiv> \<lambda>(ph,s). 
  \<forall>f \<in> fct s. (ph = P1 \<longrightarrow> f \<notin> FTest0) \<and> (ph = P2 \<longrightarrow> f \<notin> FTest)"
(*<*)
lemma no_FTest_invariant: "inv SLE66.ism no_FTest_invariant"
apply (unfold SLE66.ism_def no_FTest_invariant_def)
apply (rule invI)
apply  (auto del: conjI)
apply (erule thin_rl)
apply (erule SLE66.transs.elims)
apply (auto) 
apply (drule FTestD, fast)
apply (frule FTest0_is_FTest, drule FTest_is_Sec, drule (2) Axiom12''', fast)
apply (frule FTest_is_Sec, drule (2) Axiom12''', fast)
apply (simp add: F_NSec_def, frule FTest0_is_FTest, drule FTest_is_Sec, fast)
apply (drule FTest0_is_FTest, erule (1) FTest_not_ASec)
done
(*>*)
text {*
\vspace{-3ex}
When proving that the invariant holds, 14 of the 19 cases are trivial,
and the remaining ones require simple properties of the set @{term FTest},
and two of them require additionally \texttt{Axiom1} and \texttt{Axiom2}.
The invariant implies
*}
(*<*)
lemma P1_no_FTest: 
  "\<lbrakk>(P1,s) \<in> reach SLE66.ism; f \<in> fct s\<rbrakk> \<Longrightarrow> f \<notin> FTest0"
apply (drule no_FTest_invariant [THEN invE])
apply (auto simp: no_FTest_invariant_def)
done
(*>*)
lemma P2_no_FTest: 
  "\<lbrakk>(P2,s) \<in> reach SLE66.ism; f \<in> fct s\<rbrakk> \<Longrightarrow> f \<notin> FTest"
(*<*)
apply (drule no_FTest_invariant [THEN invE])
apply (auto simp: no_FTest_invariant_def)
done
(*>*)

text {*
\vspace{-3ex}
Exploiting this property for the case of rule \texttt{R21}, 
we can prove \texttt{FSO5} in the usual way.
This theorem states that in any sequence of transitions performed by the chip,
any attempt to execute a test function not issued by the processor manufacturer 
is refused:
*}
theorem FSO5: "\<lbrakk>ts \<in> TRuns; ((p,(ph,s)),c,(p',(ph',s'))) \<in> set ts; 
                p In = [Exec sb f]; f \<in> FTest\<rbrakk> \<Longrightarrow> 
                sb = Pmf \<or> s' = s \<and> p' Out = [No]"
(*<*)
apply (drule (1) TRunsD)
apply (safe elim!: SLE66.transs.elims del: disjCI, simp_all)
apply (fast dest: P2_no_FTest)
done
(*>*)

text {*
\vspace{-2ex}
The Isabelle proofs of all six theorems formalizing the security objectives
and the two lemmas required are well supported by Isabelle: each of them takes 
just a few steps, about half of which are automatic.
*}
(*

constdefs
  unwind :: "SLE66_state \<Rightarrow> on set \<Rightarrow> SLE66_state \<Rightarrow> bool" 
 ("_/ ~_~/ _" [61, 61, 61] 60)
 "s ~A~ t \<equiv> fst s = fst t \<and> fct (snd s) = fct (snd t) \<and>
            (\<forall>on\<in>A. val (snd s) on = val (snd t) on)"
text {* The existence of (high-level) functions is leaked to NSec. *}

lemma unwind_def2: "(ph, s) ~A~ (ph',t) = (ph = ph' \<and> fct s = fct t \<and>
                      (\<forall>fn. F fn \<in> A \<longrightarrow> valF s fn = valF t fn) \<and> 
                      (\<forall>dn. D dn \<in> A \<longrightarrow> valD s dn = valD t dn))"
by (auto simp add: unwind_def val_def split add: on.split)

lemma unwind_refl: "\<And>s. s ~A~ s" by (auto simp add: unwind_def2)

lemma unwind_ph_fct_D: "(ph, s) ~A~ (ph',t) \<Longrightarrow> ph = ph' \<and> fct s = fct t"
by (simp add: unwind_def2)

lemma unwind_NSecD:
  "\<lbrakk>(ph, s) ~- Sec~ (ph', t); on \<notin> Sec\<rbrakk> \<Longrightarrow> val s on = val t on"
by (rule_tac y = "on" in on.exhaust, auto simp: unwind_def2 val_def)

axioms noleak_output_change:
"\<lbrakk>(ph, s) ~- Sec~ (ph', t); f \<in> fct s; f \<notin> FTest\<rbrakk>
  \<Longrightarrow> output f s = output f t \<and> (ph, change f s) ~- Sec~ (ph', change f t)"

axioms noleak_Test_output:
"\<lbrakk>(ph, s) ~- Sec~ (ph', t); f \<in> fct s; f \<in> FTest\<rbrakk>
  \<Longrightarrow> positive (output f s) = positive (output f t)"

axioms allowed_single_leakage_after_Spy:
"\<lbrakk>((\<currency>(In := [Spy on]), ph, s), c, (p', Error, s')) \<in> transs;
  (ph, s) \<in> reach ism; (ph, t) \<in> reach ism; (ph, s) ~- Sec~ (ph, t); 
  on \<in> Sec; ph \<noteq> Error\<rbrakk> \<Longrightarrow> the (val s on) = the (val t on)"

theorem noleak_Sec: "\<And>s t. \<lbrakk>s \<in> reach ism; t \<in> reach ism;
  ((p,s),c,(p',s')) \<in> transs ;  s ~-Sec~ t\<rbrakk> \<Longrightarrow> \<exists>t'. 
  ((p,t),c,(p',t')) \<in> transs \<and> s' ~-Sec~ t'"
apply (clarsimp, frule unwind_ph_fct_D, clarify)
apply (rename_tac t)
apply (rule SLE66.transs.elims, erule asm_rl)
apply (safe del: conjI IntI disjE UnE insertE)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply  (fast dest: FTest0_is_FTest noleak_Test_output)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back
apply   (fast)
apply  (fast dest: FTest1_is_FTest noleak_Test_output)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (fast dest: FTest0_is_FTest noleak_Test_output)
apply (clarsimp simp add: unwind_def2)

apply (drule noleak_output_change, simp, simp, simp)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply  (fast dest: FTest1_is_FTest noleak_Test_output)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (fast dest: FTest1_is_FTest noleak_Test_output)
apply (clarsimp simp add: unwind_def2)

apply (drule noleak_output_change, simp, simp)
apply  (simp add: FTest_def, fast dest: P1_no_FTest)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply (drule noleak_output_change, simp, simp, fast dest: P2_no_FTest)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply clarsimp
apply (erule disjE)
apply  (clarsimp simp add: unwind_def2 F_NSec_def Sec_def)
apply  (fold fct_def, simp)
apply (clarsimp simp add: unwind_def2 F_ASec_def Sec_def)
apply (fold fct_def, simp)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply (frule (1) unwind_NSecD, simp)
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (fast)
apply (fast)

apply clarsimp
apply (erule disjE)
apply  ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply     (fast)
apply    (fast)
apply   (fast)
apply  (rule unwind_refl)
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply    (fast)
apply   (simp)
apply   (erule (5) allowed_single_leakage_after_Spy)
apply  (fast)
apply (rule unwind_refl)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back
apply   (fast)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back back
apply (rule unwind_refl)
done
*)
(* unused theorem noleak_Sec_determ: "\<And>s t. \<lbrakk>s ~-Sec~ t; 
  s \<in> reach ism; ((p,s),c,(p' ,s')) \<in> transs;
  t \<in> reach ism; ((p,t),c,(p'',t')) \<in> transs; 
  fst s' \<noteq> Error; fst t' \<noteq> Error\<rbrakk> \<Longrightarrow> p' = p'' \<and> s' ~-Sec~ t'"
apply (clarsimp, frule unwind_ph_fct_D, clarify)
apply (erule SLE66.transs.elims)
apply (safe del: conjI IntI disjE UnE insertE)
apply (erule_tac [!] SLE66.transs.elims, simp_all)
apply (auto del: conjI IntI disjE UnE insertE 
            dest: FTest0_is_FTest FTest1_is_FTest)

apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply fast

apply (erule (1) FTest01_disjunctE)
apply (erule (1) FTest01_disjunctE)

apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply fast

apply (erule noleak_output_change, simp, simp)

apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply clarsimp
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply fast

apply (erule noleak_output_change, simp)
apply (fast dest: P1_no_FTest FTestD)

apply (erule noleak_output_change, simp)
apply (fast dest: P2_no_FTest)

apply (erule disjE)
apply  (clarsimp simp add: unwind_def2 F_NSec_def Sec_def)
apply  (fold fct_def, simp)

apply (clarsimp simp add: unwind_def2 F_ASec_def Sec_def)
apply (fold fct_def, simp)

apply (force dest: unwind_NSecD)
done
*)
(*<*)

text {* \pagebreak *}
subsection "Nonleakage"

constdefs
  unwind :: "SLE66_state \<Rightarrow> on set \<Rightarrow> SLE66_state \<Rightarrow> bool" 
 ("_/ ~_~/ _" [61, 61, 61] 60)
 "s ~A~ t \<equiv> fst s = fst t \<and> fct (snd s) = fct (snd t) \<and>
            (\<forall>f\<in>fct (snd s). output f (snd s) = output f (snd t)) \<and>
            (\<forall>on\<in>A. val (snd s) on = val (snd t) on)"


lemma unwind_def2: "(ph, s) ~A~ (ph',t) = (ph = ph' \<and> fct s = fct t \<and>
                      (\<forall>f\<in>fct s. output f s = output f t) \<and>
                      (\<forall>fn. F fn \<in> A \<longrightarrow> valF s fn = valF t fn) \<and> 
                      (\<forall>dn. D dn \<in> A \<longrightarrow> valD s dn = valD t dn))"
(* < *)
by (auto simp add: unwind_def val_def split add: on.split)
(* > *)

lemma unwind_refl: "\<And>s. s ~A~ s" 
(* < *) 
by (auto simp add: unwind_def2)
(* > *)

lemma unwind_ph_fct_D: "(ph, s) ~A~ (ph',t) \<Longrightarrow> ph = ph' \<and> fct s = fct t"
(* < *)
by (simp add: unwind_def2)
(* > *)

lemma unwind_outputD: 
  "\<lbrakk>(ph, s) ~A~ (ph',t); f \<in> fct s\<rbrakk> \<Longrightarrow> output f s = output f t" 
(* < *)
by (simp add: unwind_def2)
(* > *)

lemma unwind_NSecD:
  "\<lbrakk>(ph, s) ~- Sec~ (ph', t); on \<notin> Sec\<rbrakk> \<Longrightarrow> val s on = val t on"
(* < *)
by (rule_tac y = "on" in on.exhaust, auto simp: unwind_def2 val_def)
(* > *)

axioms noleak_change:
"\<lbrakk>(ph, s) ~- Sec~ (ph', t); f \<in> fct s; f \<notin> FTest\<rbrakk>
  \<Longrightarrow> (ph, change f s) ~- Sec~ (ph', change f t)"

axioms disable_FTest_output_eq: 
 "\<lbrakk>(ph, s) \<in> reach ism; g \<notin> FTestX; g \<in> fct s; FTestX \<subseteq> FTest\<rbrakk> \<Longrightarrow> 
   output g (s\<lparr>valF := valF s\<lfloor>(- FTestX)\<rparr>) = output g s"

axioms load_not_PSec_function_output:
 "\<lbrakk>(P1, s) \<in> reach ism; (P1, t) \<in> reach ism; fct s = fct t; 
   f \<in> F_NSec \<or> f \<in> F_ASec \<and> f \<notin> fct t; \<forall>f\<in>fct t. output f s = output f t;
   \<forall>fn. F fn \<notin> Sec \<longrightarrow> valF s fn = valF t fn;
   \<forall>dn. D dn \<notin> Sec \<longrightarrow> valD s dn = valD t dn\<rbrakk>
 \<Longrightarrow> output f (s\<lparr>valF := valF s(f\<mapsto>v)\<rparr>) = output f (t\<lparr>valF := valF t(f\<mapsto>v)\<rparr>) \<and>
 (\<forall>fa\<in>fct t. output fa (s\<lparr>valF := valF s(f\<mapsto>v)\<rparr>) = 
               output fa (t\<lparr>valF := valF t(f\<mapsto>v)\<rparr>))"

axioms allowed_single_leakage_after_Spy:
"\<lbrakk>((\<currency>(interface.In := [Spy on]), ph, s), c, (p', Error, s')) \<in> transs;
  (ph, s) \<in> reach ism; (ph, t) \<in> reach ism; (ph, s) ~- Sec~ (ph, t); 
  on \<in> Sec; ph \<noteq> Error\<rbrakk> \<Longrightarrow> the (val s on) = the (val t on)"

theorem noleak_Sec: "\<And>s t. \<lbrakk>s \<in> reach ism; t \<in> reach ism;
  ((p,s),c,(p',s')) \<in> transs ;  s ~-Sec~ t\<rbrakk> \<Longrightarrow> \<exists>ts' t'. 
  ((p,t),c,(p',t')) \<in> transs \<and> s' ~-Sec~ t'"
(* < *)
apply (clarsimp, frule unwind_ph_fct_D, clarify)
apply (rename_tac t)
apply (rule SLE66.transs.elims, erule asm_rl)
apply (safe del: conjI IntI disjE UnE insertE)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply  (force dest: unwind_outputD)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply (rule conjI)
apply  (clarsimp)
apply  (drule (1) disable_FTest_output_eq, simp) 
apply  (fast intro: FTest0_is_FTest)
apply  (drule (1) disable_FTest_output_eq, simp) 
apply  (fast intro: FTest0_is_FTest)
apply  (clarsimp)
apply (clarsimp)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest0_is_FTest FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back
apply   (fast)
apply  (force dest: unwind_outputD)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply (rule conjI)
apply  (clarsimp)
apply  (drule (1) disable_FTest_output_eq, simp, simp) 
apply  (drule (1) disable_FTest_output_eq, simp, simp) 
apply  (clarsimp)
apply (clarsimp)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (force dest: unwind_outputD)
apply (clarsimp simp add: unwind_def2)

apply (frule unwind_outputD, simp)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (erule noleak_change, simp, simp)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply   (fast)
apply  (force dest: unwind_outputD)
apply (clarsimp simp add: unwind_def2 Sec_def)
apply (rule conjI, fold fct_def, simp)
apply (rule conjI)
apply  (clarsimp)
apply  (drule (1) disable_FTest_output_eq, simp) 
apply  (fast intro: FTest1_is_FTest)
apply  (drule (1) disable_FTest_output_eq, simp) 
apply  (fast intro: FTest1_is_FTest)
apply  (clarsimp)
apply (clarsimp)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply (rule ssubst[OF restrict_in], clarify dest!: FTest1_is_FTest FTest_is_Sec)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (force dest: unwind_outputD)
apply (clarsimp simp add: unwind_def2)

apply (frule unwind_outputD, simp)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (erule noleak_change, simp)
apply (simp add: FTest_def, fast dest: P1_no_FTest)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply (frule unwind_outputD, simp)
apply clarsimp
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (erule noleak_change, simp, fast dest: P2_no_FTest)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (clarsimp simp add: unwind_def2)
apply (fold fct_def, simp)
apply (erule (6) load_not_PSec_function_output)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros)
apply  (fast)
apply (fast)

apply (frule (1) unwind_NSecD, simp)
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply   (fast)
apply  (fast)
apply (fast)

apply clarsimp
apply (erule disjE)
apply  ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply    (fast)
apply   (fast)
apply  (fast)
apply (rule unwind_refl)
apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) 
apply    (fast)
apply   (simp)
apply   (erule (5) allowed_single_leakage_after_Spy)
apply  (fast)
apply (rule unwind_refl)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back
apply   (fast)
apply  (fast)
apply (fast)

apply ((rule exI)+, rule conjI, rule SLE66.transs.intros) back back
apply (rule unwind_refl)
done

(*>*)
end
