(*  Title:      ISM/dISM.thy
    Id:         $Id: dISM.thy,v 1.5 2003/03/07 21:07:23 dvo Exp $
    Author:     David von Oheimb
    License:    (C) 2002 Siemens AG; any non-commercial use is granted

*)

header {* dynamic ISMs *}

theory dISM
imports ISM
begin
record ('i, 'p) dstate = 
  running :: "'i set"
  enabled :: "'p set"
  owned   :: "('i \<Rightarrow> 'p set)"  --{* also used as specification of allowed inputs *}

datatype ('i, 'p) dcmd 
= Run     'i 
| Stop    'i
| New     'p 
| Convey  'p 'i  --{* port name and name of new owner *}
| Enable  'p 
| Disable 'p

consts
  dstate_trans :: "'i set \<Rightarrow> ('i, ('i, 'p) dcmd, ('i, 'p) dstate) g_trans"
(*<*)
syntax (xsymbols) 
 "@dstate_trans" :: 
  "('i, 'p) dstate \<Rightarrow> 'i \<Rightarrow> 'i set \<Rightarrow> ('i, 'p) dcmd \<Rightarrow>  ('i, 'p) dstate \<Rightarrow> bool"
                                     ("_ \<midarrow>_:_:_\<rightarrow> _"  [61, 61, 61, 61, 61] 60)
 "@dstate_trans*":: 
  "('i, 'p) dstate \<Rightarrow> 'i \<Rightarrow> 'i set \<Rightarrow> ('i, 'p) dcmd list\<Rightarrow> ('i,'p)dstate\<Rightarrow> bool"
                                     ("_ \<midarrow>_:_:_\<rightarrow>* _" [61, 61, 61, 61, 61] 60)
translations
 "di \<midarrow>i:js:dcmd'\<rightarrow>  di'" \<rightleftharpoons> "(di, dcmd', di') \<in>           dstate_trans js i"
 "di \<midarrow>i:js:dcmds\<rightarrow>* di'" \<rightleftharpoons> "(di, dcmds, di') \<in> fold_rel (dstate_trans js i)"
(*>*)
inductive "dstate_trans js i" intros
--{* only a current owner may receive from, enable, disable, and convey a port,
     \\\hspace*{2.5ex} otherwise transition is blocked *}
Run:  "\<lbrakk>j \<notin> running di; j \<in> js\<rbrakk> \<Longrightarrow> di \<midarrow>i:js:Run  j\<rightarrow> di\<lparr>running:=running di\<union>{j}\<rparr>"
Stop: "\<lbrakk>j \<in> running di; j \<in> js\<rbrakk> \<Longrightarrow> di \<midarrow>i:js:Stop j\<rightarrow> di\<lparr>running:=running di-{j}\<rparr>"
Enable:  "\<lbrakk>p \<notin> enabled di; p \<in> owned di i\<rbrakk> \<Longrightarrow> 
                               di \<midarrow>i:js:Enable p\<rightarrow> di\<lparr>enabled:=enabled di\<union>{p}\<rparr>"
Disable: "\<lbrakk>p \<in> enabled di; p \<in> owned di i\<rbrakk> \<Longrightarrow> 
                              di \<midarrow>i:js:Disable p\<rightarrow> di\<lparr>enabled:=enabled di-{p}\<rparr>"
--{* newly created ports are initially owned by their creator and are (implicitly) enabled. *}
New:     "\<lbrakk>p \<notin> (\<Union>i. owned di i)\<rbrakk> \<Longrightarrow> 
                di \<midarrow>i:js:New p\<rightarrow> di\<lparr>owned:=(owned di)(i:=owned di i\<union>{p})\<rparr>"
Convey:  "\<lbrakk>p \<in> owned di i; p \<notin> owned di j; j \<in> js\<rbrakk> \<Longrightarrow> 
                di \<midarrow>i:js:Convey p j\<rightarrow> di\<lparr>owned:=(owned di)(j:=owned di j\<union>{p},
                                                          \<spacespace>i:=owned di i-{p})\<rparr>"

constdefs
  d_trans :: "'i set \<Rightarrow> ('i, ('i, 'p) dcmd list, ('i, 'p) dstate) g_trans"
 "d_trans js \<equiv> \<lambda>i. {(di,dcmds,di'). i \<in> running di \<and> di \<midarrow>i:js:dcmds\<rightarrow>* di'}"

constdefs
  init_dstate :: "('i, ('c, 'p, 'm, 's) ism) family \<Rightarrow> 'i set \<Rightarrow> ('i, 'p) dstate"
 "init_dstate \<equiv> \<lambda>(A,I) run. \<lparr>running = run, enabled = UNIV, 
                             owned = (\<lambda>i. if i \<in> I then inputs (A i) else {})\<rparr>"

  set_in_out :: "('i, ('c, 'p, 'm, 's) ism) family \<Rightarrow> ('i, 'p) dstate \<Rightarrow> 
                 ('i, ('c, 'p, 'm, 's) ism) family"
 "set_in_out \<equiv> \<lambda>(A,I) di. (\<lambda>i. (A i)\<lparr>inputs := owned di i, outputs := 
     outputs (A i) \<inter> enabled di \<inter> (\<Union>j\<in>running di. owned di j)\<rparr>,I)"

  d_comp_runs :: "('i, (('i, 'p) dcmd list, 'p, 'm, 's) ism) family \<Rightarrow> 'i set \<Rightarrow>
           ((('p, 'm, ('i, 'p) dstate \<times> ('i \<Rightarrow> 's)) conf) list) set"
 "d_comp_runs AI run \<equiv> comp_runs (set_in_out AI) (init_dstate AI run) (d_trans UNIV)" 

--{* transitions only for running dISMs; marking non-existing dISMs as running is harmless *}
--{* input only on ports currently owned (which may be enabled or not), otherwise blocked *}
--{* output only to enabled ports of running ISMs, otherwise transition is blocked *}

(*<*)
lemma wf_comp_set_in_out_init_dstateI:
"\<And>AI. wf_comp AI \<Longrightarrow> wf_comp (set_in_out AI (init_dstate AI run))"
by (force simp add: set_in_out_def init_dstate_def wf_comp_def)

lemma snd_set_in_out_eq [simp]:
"\<And>AI. snd (set_in_out AI di) = snd AI"
by (auto simp add: set_in_out_def)

lemma d_trans_preserves_wf_comp_set_in_out [rule_format]:
"\<And>AI. \<lbrakk>i \<in> snd AI; (g1, c, g2) \<in> d_trans js i\<rbrakk> \<Longrightarrow> 
  wf_comp (set_in_out AI g1) \<longrightarrow> wf_comp (set_in_out AI g2)"
apply (auto simp add: d_trans_def set_in_out_def init_dstate_def wf_comp_def  
            del: impI)
apply (erule fold_rel.induct)
apply  simp_all
apply (rule, erule mp)
apply (rotate_tac -1, rule mp, erule_tac [2] asm_rl, erule thin_rl)
apply (erule dstate_trans.induct, auto)
done
(*>*)
lemma d_comp_runs_preserves_wf_comp: 
 "wf_comp AI \<Longrightarrow> cs \<in> d_comp_runs AI run \<Longrightarrow> 
  \<forall>(b,di,s) \<in> set cs. wf_comp (set_in_out AI di)"
(*<*)
apply (unfold d_comp_runs_def, erule comp_runs_global_induct, simp_all)
apply  (erule wf_comp_set_in_out_init_dstateI)
apply (erule (2) d_trans_preserves_wf_comp_set_in_out)
done
(*>*)

end
